'use strict';

var lodash = require('lodash');
var lokka = require('lokka');
var lokkaTransportHttp = require('lokka-transport-http');
var redux = require('redux');

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) {
  return typeof obj;
} : function (obj) {
  return obj && typeof Symbol === "function" && obj.constructor === Symbol ? "symbol" : typeof obj;
};

/*
  ROULADE ACTION TYPES
  This may deserve its own file
*/

// Mutation Actions
var ROULADE_MUTATION_REQUEST = 'ROULADE_MUTATION_REQUEST';
var ROULADE_MUTATION_SUCCESS = 'ROULADE_MUTATION_SUCCESS';
var ROULADE_MUTATION_ERROR = 'ROULADE_MUTATION_ERROR';

// Query Actions
var ROULADE_QUERY_REQUEST = 'ROULADE_QUERY_REQUEST';
var ROULADE_QUERY_SUCCESS = 'ROULADE_QUERY_SUCCESS';
var ROULADE_QUERY_ERROR = 'ROULADE_QUERY_ERROR';

// Query Actions
function query(root, vars, dataSpec) {
  var options = arguments.length <= 3 || arguments[3] === undefined ? {} : arguments[3];

  // Potential options include path, action, and noload.
  // Fill in default for path.  action and noload should remain undefined if undefined.
  var action = options.action;
  var noload = options.noload;

  var path = options.path ? options.path : 'search';

  var client = createClient(options);

  return function (dispatch) {
    // handle loading state unless noload option specified
    if (!noload) {
      dispatch(queryRequest(root, vars, dataSpec));
    }

    var q = void 0;
    if (typeof root === 'string') {
      q = root;
    } else {
      // FIXME: Assert about the nature of `root`. We expect the result of a
      // roulade.QL template string, but we don't make sure that's what we've
      // got.
      q = root.pre + dataSpec.fragment;
      if (root.post) {
        q += root.post;
      }
    }

    client.query(q, vars).then(function (response) {
      // Set path to 'search', the default, if undefined
      var responsePath = lodash.isUndefined(path) ? 'search' : path;
      response = response[responsePath];
      /* TODO - Add docs for this.
        if you pass an action to roulade's query action, it calls that action
        instead of updating the state itself. So if that action is present, we call
        both querySuccess as well as whatever action was passed to query() passing
        both the same arguments so users can do whatever they want w/ the query
        response.
      */
      if (action) {
        var customActionOnSuccess = function customActionOnSuccess() {
          dispatch(querySuccess(response, root, vars, dataSpec, action, noload));
          dispatch(action(response, root, vars, dataSpec));
        };
        return customActionOnSuccess();
      } else {
        return dispatch(querySuccess(response, root, vars, dataSpec));
      }
    }, function (error) {
      console.log('GraphQL Query Failed: ', error);
      return dispatch(queryError(error, root, vars, dataSpec));
    });
  };
}

function queryRequest(root, vars, fragment) {
  if (root === undefined) {
    throw new TypeError('`root` is required');
  }
  if (vars === undefined) {
    throw new TypeError('`vars` is required');
  }
  if (fragment === undefined) {
    throw new TypeError('`fragment` is required');
  }

  return {
    type: ROULADE_QUERY_REQUEST,
    root: root,
    vars: vars,
    fragment: fragment
  };
}

function querySuccess(response, root, vars, fragment, action, noload) {
  if (response === undefined) {
    throw new TypeError('`response` is required');
  }
  if (root === undefined) {
    throw new TypeError('`root` is required');
  }
  if (vars === undefined) {
    throw new TypeError('`vars` is required');
  }
  if (fragment === undefined) {
    throw new TypeError('`fragment` is required');
  }
  if (action) {
    return {
      type: ROULADE_QUERY_SUCCESS,
      response: response,
      root: root,
      vars: vars,
      fragment: fragment,
      action: action,
      noload: noload
    };
  }

  return {
    type: ROULADE_QUERY_SUCCESS,
    response: response,
    root: root,
    vars: vars,
    fragment: fragment,
    action: action
  };
}

function queryError(error, root, vars, fragment) {
  if (error === undefined) {
    throw new TypeError('`error` is required');
  }
  if (root === undefined) {
    throw new TypeError('`root` is required');
  }
  if (vars === undefined) {
    throw new TypeError('`vars` is required');
  }
  if (fragment === undefined) {
    throw new TypeError('`fragment` is required');
  }

  return {
    type: ROULADE_QUERY_ERROR,
    error: error,
    root: root,
    vars: vars,
    fragment: fragment
  };
}

// Mutation Actions
function mutate(name, fragment, vars) {
  var options = arguments.length <= 3 || arguments[3] === undefined ? {} : arguments[3];
  var noload = options.noload;
  var request = options.request;
  var error = options.error;
  var success = options.success;


  var client = createClient(options);
  var mutation = '{\n    ' + name + '(' + formatVarObj(vars) + ')\n    ' + fragment + '\n  }';

  // this gives each mutation a unique ID for loading state purposes, however its
  // not known to the component that cares about it, so... blerg. See reducer for
  // more details on mutation loading state question.
  var uniqueName = lodash.uniqueId(name + '_');
  function formatVarObj(varObj) {
    // stringify & remove quotes from only 1st level keys
    var varString = JSON.stringify(varObj).replace(/\"([^(\")"]+)\":/g, "$1:");
    // return string after trimming braces
    return varString.substr(1, varString.length - 2);
  }

  return function (dispatch) {
    // handle loading state unless noLoad option specified
    !noload ? dispatch(mutationRequest(uniqueName, fragment, vars, options)) : null;
    client.mutate(mutation).then(function (response) {
      var result = success ? dispatch(success(response, name, fragment, vars, options)) && dispatch(mutationSuccess(response, uniqueName, fragment, vars, options)) : dispatch(mutationSuccess(response, uniqueName, fragment, vars, options));
      return result;
    }, function (error) {
      return dispatch(mutationError(error, uniqueName, fragment, vars, options));
    });
  };
}

// Given an option object, construct an appropriate Lokka client.
// TODO: Give consumers more control over this. Maybe they won't want to do
// HTTP transport.
function createClient(options) {
  var host = options.host || '';

  // TODO: The change for headers and host should be replicated in Backstage.
  // Or we should finally pull Roulade out.
  var headers = options.token ? {
    authorization: 'Token token=\'' + options.token + '\''
  } : {};

  return new lokka.Lokka({
    transport: new lokkaTransportHttp.Transport(host + '/graphql', { headers: headers })
  });
}

function mutationRequest(name, fragment, vars, options) {
  if (name === undefined || typeof name !== 'string') {
    throw new TypeError('`name` is required and must be a string');
  }
  if (fragment === undefined || typeof fragment !== 'string') {
    throw new TypeError('`fragment` is required and must be a string');
  }
  if (vars === undefined || (typeof vars === 'undefined' ? 'undefined' : _typeof(vars)) !== 'object') {
    throw new TypeError('`vars` is required and must be an object');
  }
  return {
    type: ROULADE_MUTATION_REQUEST,
    name: name, fragment: fragment, vars: vars, options: options
  };
}
function mutationSuccess(response, name, fragment, vars, options) {
  if (name === undefined || typeof name !== 'string') {
    throw new TypeError('`name` is required and must be a string');
  }
  if (fragment === undefined || typeof fragment !== 'string') {
    throw new TypeError('`fragment` is required and must be a string');
  }
  if (vars === undefined || (typeof vars === 'undefined' ? 'undefined' : _typeof(vars)) !== 'object') {
    throw new TypeError('`vars` is required and must be an object');
  }
  return {
    type: ROULADE_MUTATION_SUCCESS,
    response: response, name: name, fragment: fragment, vars: vars, options: options
  };
}
function mutationError(error, name, fragment, vars, options) {
  if (name === undefined || typeof name !== 'string') {
    throw new TypeError('`name` is required and must be a string');
  }
  if (name === undefined || typeof name !== 'string') {
    throw new TypeError('`name` is required and must be a string');
  }
  if (fragment === undefined || typeof fragment !== 'string') {
    throw new TypeError('`fragment` is required and must be a string');
  }
  if (vars === undefined || (typeof vars === 'undefined' ? 'undefined' : _typeof(vars)) !== 'object') {
    throw new TypeError('`vars` is required and must be an object');
  }
  return {
    type: ROULADE_MUTATION_ERROR,
    error: error, name: name, fragment: fragment, vars: vars, options: options
  };
}



var actions = Object.freeze({
  ROULADE_MUTATION_REQUEST: ROULADE_MUTATION_REQUEST,
  ROULADE_MUTATION_SUCCESS: ROULADE_MUTATION_SUCCESS,
  ROULADE_MUTATION_ERROR: ROULADE_MUTATION_ERROR,
  ROULADE_QUERY_REQUEST: ROULADE_QUERY_REQUEST,
  ROULADE_QUERY_SUCCESS: ROULADE_QUERY_SUCCESS,
  ROULADE_QUERY_ERROR: ROULADE_QUERY_ERROR,
  query: query,
  queryRequest: queryRequest,
  querySuccess: querySuccess,
  queryError: queryError,
  mutate: mutate,
  mutationRequest: mutationRequest,
  mutationSuccess: mutationSuccess,
  mutationError: mutationError
});

// This generates a cursor for a nonexistent field in application state to handle
// loading status. Its injected in the top level of the cursor (side by side w/
// where the query results end up). It adds 'Status' to the same name. IE, a Query
// for results would end up with:
// results: {
//   resultsStatus: 'loading' || 'loaded'
// }
// Its up to the components to consume this status as they will, but the status
// of a component's query will be made available to the component.
function generateLoadingCursor(type, cursor) {
  var subcursorArray = cursor.match("^[^.]*");
  return subcursorArray[0] + '.' + subcursorArray[0] + type + 'Status';
}

// Roulade Reducer
function rouladeReducer(state, action) {
  var oldState = lodash.cloneDeep(state);
  switch (action.type) {

    // Query handlers
    case 'ROULADE_QUERY_REQUEST':
      var loadingQueryCursor = generateLoadingCursor('Query', action.fragment.cursor);
      return lodash.set(oldState, loadingQueryCursor, 'loading');
    case 'ROULADE_QUERY_ERROR':
      var errorErrorCursor = generateLoadingCursor('Query', action.fragment.cursor);
      return lodash.set(oldState, errorErrorCursor, 'error');
    case 'ROULADE_QUERY_SUCCESS':
      if (!action.noload) {
        if (action.action) {
          var loadedQueryCursor = generateLoadingCursor('Query', action.fragment.cursor);
          return lodash.set(oldState, loadedQueryCursor, 'loaded');
        } else {
          var stateWithQueryResponse = lodash.set(oldState, action.fragment.cursor, action.response);
          var _loadedQueryCursor = generateLoadingCursor('Query', action.fragment.cursor);
          return lodash.set(stateWithQueryResponse, _loadedQueryCursor, 'loaded');
        }
      } else {
        return oldState;
      }
      break;

    // Mutation Handlers
    // So, do we need loading state for the mutations? I'm not sure how a component
    // would know the particular instance of mutation state would apply to it,
    // especially considering it gets a random ID attached in roulade actions.
    // Specifically, do we need to handle loading state at all w/ mutations?
    case 'ROULADE_MUTATION_REQUEST':
      // const loadingMutationCursor = generateLoadingCursor('results.mutation', action.name)
      // return set(oldState, loadingMutationCursor, 'loading')
      return state;
    case 'ROULADE_MUTATION_ERROR':
      // const errorMutationCursor = generateLoadingCursor('results.mutation', action.name)
      // return set(oldState, errorMutationCursor, 'error')
      return state;
    case 'ROULADE_MUTATION_SUCCESS':
      // if (!action.noload) {
      //   if (action.success) {
      //     const loadedMutationCursor = generateLoadingCursor('results.mutation', action.name)
      //     // Change this to remove loading state
      //     return unset(oldState, loadedMutationCursor)
      //   }
      // }
      return state;
    default:
      return state;
  }
}

// If user wants to replace redux's combineReducers fn
var combineReducers$1 = function combineReducers(reducers) {
  var combinedReducers = redux.combineReducers(reducers);
  return function (state, action) {
    var updatedAppState = combinedReducers(state, action);
    return rouladeReducer(updatedAppState, action);
  };
};

// Exports
function wrapper(component, data) {
  component.helloWorld = function () {
    console.log('added helloWorld method to component');
  };
  // Unused, mostly for logging/testing purposes
  component.exposeCursor = function () {
    return data.cursor;
  };
  component.merge = function (obj1, obj2) {
    return lodash.merge(obj1, obj2);
  };
  // used when calling actions
  component.exposeData = function () {
    return {
      cursor: data.cursor,
      fragment: data.fragment
    };
  };
  // used when composing fragments
  component.exposeFragment = function () {
    return data.fragment;
  };
  // expose mutation
  component.exposeMutations = function () {
    return {
      mutations: data.mutations
    };
  };
  // get specific mutation
  component.getMutation = function (mutation) {
    return data.mutations[mutation];
  };

  return component;
}

var BODY = Symbol('body marker');

// This is a tagged template string function that accepts input of the form:
//
//     roulade.QL`query {${BODY}}`
//
// And produces an object like:
//
//     {
//       pre: 'query {',
//       post: '}',
//     }
//
// Everything up to the first encountered BODY is placed in `pre`, all
// subsequent content is in `post`.
function QL(strings) {
  var pre = '';
  var post = null;

  for (var _len = arguments.length, values = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
    values[_key - 1] = arguments[_key];
  }

  for (var i = 0; i < strings.length; i++) {
    if (post === null) {
      pre += strings[i];
    } else {
      post += strings[i];
    }

    if (i < values.length) {
      if (values[i] === BODY) {
        post = '';
      } else {
        pre += values[i];
      }
    }
  }
  return { pre: pre, post: post };
}

exports.actions = actions;
exports.reducer = rouladeReducer;
exports.combineReducers = combineReducers$1;
exports.wrapper = wrapper;
exports.QL = QL;
exports.BODY = BODY;